import alembic.autogenerate

from .enum_type import EnumType
from . import alembic_ops


def get_declared_enums(metadata):
    types = set(
        column.type
        for table in metadata.tables.values()
        for column in table.columns
        if isinstance(column.type, EnumType)
    )
    return {
        typ.__enum__.__tablename__: frozenset(typ.__enum__.__enum__.__members__)
        for typ in types
    }


def is_table_present(tablename, connection):
    try:
        connection.execute("SELECT * FROM {} LIMIT 0;".format(tablename))
    except:
        return False
    else:
        return True


@alembic.autogenerate.comparators.dispatch_for("schema")
def compare_enums(autogen_context, upgrade_ops, schema_names):
    for schema in schema_names:
        default = autogen_context.dialect.default_schema_name
        if schema is None:
            schema = default

        enums = get_declared_enums(autogen_context.metadata)
        for table, values in enums.items():
            if is_table_present(table, autogen_context.connection):
                items = {
                    r[0]
                    for r in autogen_context.connection.execute(
                        "SELECT item_id FROM {}".format(table)
                    )
                }
                to_add = values - items
                to_remove = items - values
                if to_add:
                    upgrade_ops.ops.append(
                        alembic_ops.EnumInsertOp(table, list(to_add))
                    )
                if to_remove:
                    upgrade_ops.ops.append(
                        alembic_ops.EnumDeleteOp(table, list(to_remove))
                    )
            else:
                upgrade_ops.ops.append(alembic_ops.EnumInsertOp(table, list(values)))



from .enum_table import EnumTable
from .enum_type import EnumType

try:
    import alembic
except ModuleNotFoundError:
    pass  # Expected exception
else:
    from .alembic_ops import EnumInsertOp, EnumDeleteOp
    from .alembic_autogen import compare_enums



import sqlalchemy.types as types


__all__ = ["EnumType"]


class EnumType(types.TypeDecorator):
    impl = types.String

    def __init__(self, enum_table=None, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.__enum__ = enum_table

    def process_bind_param(self, value, dialect):
        if value is None:
            return None
        return value.name

    def process_result_value(self, value, dialect):
        if value is None:
            return None
        return self.__enum__.__enum__[value]



from alembic.operations import Operations, MigrateOperation
import alembic.autogenerate.render

__all__ = ["EnumInsertOp", "EnumDeleteOp"]


@Operations.register_operation("enum_insert")
class EnumInsertOp(MigrateOperation):
    def __init__(self, tablename, data=[]):
        self.tablename = tablename
        self.data = data

    @classmethod
    def enum_insert(cls, operations, tablename, data=[]):
        op = cls(tablename, data)
        return operations.invoke(op)

    def reverse(self):
        return EnumDeleteOp(self.tablename, self.data)


@Operations.register_operation("enum_delete")
class EnumDeleteOp(MigrateOperation):
    def __init__(self, tablename, data=[]):
        self.tablename = tablename
        self.data = data

    @classmethod
    def enum_delete(cls, operations, tablename, data=[]):
        op = cls(tablename, data)
        return operations.invoke(op)

    def reverse(self):
        return EnumInsertOp(self.tablename, self.data)


@Operations.implementation_for(EnumInsertOp)
def insert(operations, operation):
    if not operation.data:
        return
    values = ", ".join("('" + v + "')" for v in operation.data)
    txt = "INSERT INTO {tn} (item_id) VALUES {vl}".format(
        tn=operation.tablename, vl=values
    )
    operations.execute(txt)


@Operations.implementation_for(EnumDeleteOp)
def delete(operations, operation):
    if not operation.data:
        return
    values = ", ".join("'" + v + "'" for v in operation.data)
    txt = "DELETE FROM {tn} WHERE item_id IN ({vl})".format(
        tn=operation.tablename, vl=values
    )
    operations.execute(txt)


@alembic.autogenerate.render.renderers.dispatch_for(EnumInsertOp)
def render_sync_enum_value_op(autogen_context, op):
    return 'op.enum_insert("{}", {!r})'.format(op.tablename, op.data)


@alembic.autogenerate.render.renderers.dispatch_for(EnumDeleteOp)
def render_sync_enum_value_op(autogen_context, op):
    return 'op.enum_delete("{}", {!r})'.format(op.tablename, op.data)



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['enumtables']

package_data = \
{'': ['*']}

install_requires = \
['sqlalchemy>=1.3.18,<2.0.0']

setup_kwargs = {
    'name': 'gino-enum-tables',
    'version': '1.0.5',
    'description': 'Support for storing enum values in a table instead of a type',
    'long_description': 'GINO Enumeration Tables\n====================================\n\nThis is a port of `SqlAlchemy Enum Tables <https://git.heptacle.fr/neshiraini/sqlalchemy-enum-tables>`__ to support `GINO orm <https://github.com/python-gino/gino>`__\n\nSQLAlchemy has built-in ``enum.Enum`` support,\nvia its column type ``sqlalchemy.Enum``.\nHowever, this type relies either on the backend\'s enum type,\nor on a check constraints. Both of these are immutable objects,\nwhich are a pain in the butt to modify\n(only PostgreSQL supports adding values to an enum type,\nand even then it doesn\'t support removing them).\n\nAnother often-used pattern to support enums in a database\nis via a dedicated table that reflects the enum values.\nThis requires updating the table everytime the enum is modified,\nbut doing so is much simpler than replacing a type.\n\nThis package allows you to create the enum table,\nand columns referencing that table, directly from\na Python enum class. It also interfaces with Alembic\nto automatically add ``INSERT`` and ``DELETE`` statements\nto your autogenerated migration scripts.\n\nWhen to use\n-----------\n\n1. Only works with Pythons\'s enumeration classes,\n   or at least one with a behavior similar to ``enum.Enum``.\n   Does not work with collections of arbitrary entries.\n2. Better used for frequently updated enumeration classes.\n3. **Do not** use with another package that provides\n   ``op.enum_insert`` and ``op.enum_delete`` operations in Alembic.\n\nInstallation\n------------\n::\n\n    pip install gino-enum-tables\n\nHow to use with GINO\n----------------------\n\n::\n\n    import enum\n\n    # Gino instance\n    from . import db\n\n    import enumtables as et\n\n    # Create the Python enumeration class\n    class MyEnum(enum.Enum):\n        HELLO = "HELLO"\n        WORLD = "WORLD"\n\n    # Create the enumeration table\n    # Pass your enum class and the GINO base model to enumtables.EnumTable\n    MyEnumTable = et.EnumTable(MyEnum, db.Model)\n\n    # Create a model class that uses the enum\n    class MyModel(Base):\n        __tablename__ = "my_model"\n        # Pass the enum table (not the enum class) to enumtables.EnumType\n        # Add the foreign key to the enum table\n        enum_value = db.Column(et.EnumType(MyEnumTable), db.ForeignKey(\'my_enum_table.item_id\') primary_key = True)\n\n        # When valued (on an instance of MyModel), enum_value will be an instance of MyEnum.\n\nFirst, the ``EnumTable`` factory takes the enum class and the base GINO model class\nto create the actual ORM class. Then this ORM class is passed to the ``EnumType`` custom type class, along with a ForeignKey to the enum table,\nto create the SQLAlchemy column linked to the enum table.\nThe column behaves just as if it had SqlAlchemy\'s own ``Enum`` type.\n\nNote that the ForeignKey points to \'my_enum_table.item_id\'. If the tablename argument isn\'t passed to EnumTable, the table name is created from the Enum name by doing a camelCase to snake_case conversion. \'item_id\' is the primary key of the enum table (my_enum_table) holding enum values.\n\nOn the implementation side, ``EnumTable`` is not a class,\nit\'s a factory function that performs Python black magic\nto create a subclass of the declarative base, and set it up to be a DB table\ncontaining the enum items (actually it just has one column ``item_id`` of type String).\n\nHow to use with Alembic\n-----------------------\n\nFirst add::\n\n    import enumtables\n\nat the begining of your ``env.py`` file,\nthen add the same line in the imports of your ``script.py.mako`` file.\nThe package uses Alembic\'s standard hooks to take care of migration generation.\n\nDon\'t forget to review the migrations afterwards.\nEspecially make sure that, if the table did not exist before,\nthe ``op.enum_insert`` commands are located *after* the corresponding ``op.create_table`` command.\n\nOther uses\n-----------\n\nUsing the enum table class directly\n^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^\n\nThe enum table class behaves like any SqlAlchemy ORM class::\n\n    enum_query = session.query(MyEnumTable)\n    result = enum_query.first()\n\n    # The column item_id stores the name of the enum item as a string\n    enum_name = result.item_id\n\nAdding more columns to the enum tables\n^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^\n\nAny keyword argument passed to the ``EnumTable`` factory becomes a member of the table class.\nWhich means, you can pass anything (like a column) exactly as you would defined a usual ORM class::\n\n    BetterEnumTable = et.EnumTable(\n        MyEnum,\n\t\tBase,\n\n        # tablename is turned into __tablename__\n        tablename = "better_enum",\n\n        # Let\'s add a new column!\n        order = sa.Column(sa.Integer, nullable = False),\n\n        # And since it\'s an ordering number, let\'s make it unique too.\n        __table_args__ = (\n            sa.UniqueConstraint(\'order\'),\n        ),\n    )\n',
    'author': 'Gazorby',
    'author_email': 'gazorby@pm.me',
    'maintainer': None,
    'maintainer_email': None,
    'url': None,
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'python_requires': '>=3.8,<4.0',
}


setup(**setup_kwargs)



import re
import sqlalchemy as sa

__all__ = ["EnumTable"]


def convert_case(name):
    s_1 = re.sub("(.)([A-Z][a-z]+)", r"\1_\2", name)
    return re.sub("([a-z0-9])([A-Z])", r"\1_\2", s_1).lower()


def EnumTable(enum, base_gino_model, name=None, tablename=None, doc=None, **kwargs):
    """
	Create a table for a Python enumeration

	Parameters
	----------
	enum : subclass of enum.Enum
		The enumeration for which the table will be created.
	base_gino_model : GINO base model class
		The declarative base to use to create the enum class.
	name : str
		The name of the class to create, defaults to the name of the enum class with ``Table`` appended.
		For an enum class ``SpamEnum``, the default name would be ``SpamEnumTable``.
	tablename : str
		The name of the table in the database.
		Defaults to the name of the enum class converted to snake_case.
		WIth the example above, it would be ``spam_enum``.
	doc : str
		A docstring to add to the class.
		If left to ``None``, no docstring will be added.

	Additional keyword parameters become members of the class.

	Returns
	-------
	table : Model
		The table class, an subclass of ``Model``.
		The original enum is accissible as the ``__enum__`` attribute.
		It has one column, ``item_id`` of type String.
	"""

    typename = name if name else (enum.__name__ + "Table")
    namespace = {
        "__tablename__": tablename if tablename else convert_case(enum.__name__),
        "__enum__": enum,
        "item_id": sa.Column(sa.String, primary_key=True),
        **kwargs,
    }
    if doc:
        namespace["__doc__"] = doc
    return base_gino_model.__class__(typename, (base_gino_model,), namespace)
